"""Stream walkers are the basic data retrieval mechanism in sensor graph."""

from builtins import range
from iotile.core.exceptions import ArgumentError, InternalError
from iotile.core.hw.reports import IOTileReading
from iotile.sg.exceptions import StreamEmptyError, UnresolvedIdentifierError
from iotile.sg.stream import DataStream, DataStreamSelector


class StreamWalker(object):
    """An object that walks through the SensorLog and returns matching readings.

    Args:
        selector (DataStreamSelector): The selector for the stream(s) that we
            are walking.
    """

    def __init__(self, selector):
        self.selector = selector

    def matches(self, stream):
        """Check if a stream matches this walker.

        Args:
            stream (DataStream): The stream to check

        Returns:
            bool: True if there is match, otherwise False
        """

        return self.selector.matches(stream)

    @property
    def buffered(self):
        """Whether this stream walker is backed by actual persistent storage (True)."""

        return self.selector.buffered


class BufferedStreamWalker(StreamWalker):
    """A stream walker backed by a storage buffer.

    Args:
        selector (DataStreamSelector): The selector for the streams
            that we are walking
        engine (StorageEngine): The storage engine backing us up
        skip_all (bool): Whether to start at the beginning of the data
            or to skip everything and start at the end.  Defaults
            to skipping everything.
    """

    def __init__(self, selector, engine, skip_all=True):
        super(BufferedStreamWalker, self).__init__(selector)
        self.engine = engine

        storage, streaming = self.engine.count()

        self._count = 0
        self.offset = 0
        if not skip_all:
            self._count = engine.count_matching(selector)

        if selector.output:
            if skip_all:
                self.offset = streaming

            self.storage_type = u'streaming'
        else:
            if skip_all:
                self.offset = storage

            self.storage_type = u'storage'

    def dump(self):
        """Dump the state of this stream walker.

        Returns:
            dict: The serialized state.
        """

        return {
            u'selector': str(self.selector),
            u'offset': self.offset
        }

    def restore(self, state):
        """Restore a previous state of this stream walker.

        Raises:
            ArgumentError: If the state refers to a different selector or the
                offset is invalid.
        """

        selector = DataStreamSelector.FromString(state.get(u'selector'))
        if selector != self.selector:
            raise ArgumentError("Attempted to restore a BufferedStreamWalker with a different selector",
                                selector=self.selector, serialized_data=state)

        self.seek(state.get(u'offset'), target="offset")

    def count(self):
        """Return the count of available readings in this stream walker."""

        return self._count

    def pop(self):
        """Pop a reading off of this stream and return it."""

        if self._count == 0:
            raise StreamEmptyError("Pop called on buffered stream walker without any data", selector=self.selector)

        while True:
            curr = self.engine.get(self.storage_type, self.offset)
            self.offset += 1

            stream = DataStream.FromEncoded(curr.stream)
            if self.matches(stream):
                self._count -= 1
                return curr

    def seek(self, value, target="offset"):
        """Seek this stream to a specific offset or reading id.

        There are two modes of use.  You can seek to a specific reading id,
        which means the walker will be positioned exactly at the reading
        pointed to by the reading ID.  If the reading id cannot be found
        an exception will be raised.  The reading id can be found but corresponds
        to a reading that is not selected by this walker, the walker will
        be moved to point at the first reading after that reading and False
        will be returned.

        If target=="offset", the walker will be positioned at the specified
        offset in the sensor log. It will also update the count of available
        readings based on that new location so that the count remains correct.

        The offset does not need to correspond to a reading selected by this
        walker.  If offset does not point to a selected reading, the effective
        behavior will be as if the walker pointed to the next selected reading
        after `offset`.

        Args:
            value (int): The identifier to seek, either an offset or a
                reading id.
            target (str): The type of thing to seek.  Can be offset or id.
                If id is given, then a reading with the given ID will be
                searched for.  If offset is given then the walker will
                be positioned at the given offset.

        Returns:
            bool: True if an exact match was found, False otherwise.

            An exact match means that the offset or reading ID existed and
            corresponded to a reading selected by this walker.

            An inexact match means that the offset or reading ID existed but
            corresponded to reading that was not selected by this walker.

            If the offset or reading ID could not be found an Exception is
            thrown instead.

        Raises:
            ArgumentError: target is an invalid string, must be offset or
                id.
            UnresolvedIdentifierError: the desired offset or reading id
                could not be found.
        """

        if target not in (u'offset', u'id'):
            raise ArgumentError("You must specify target as either offset or id", target=target)

        if target == u'offset':
            self._verify_offset(value)
            self.offset = value
        else:
            self.offset = self._find_id(value)

        self._count = self.engine.count_matching(self.selector, offset=self.offset)

        curr = self.engine.get(self.storage_type, self.offset)
        return self.matches(DataStream.FromEncoded(curr.stream))

    def _find_id(self, reading_id):
        shared = [None]

        def _id_searcher(i, reading):
            """Find the offset of the first reading with the given reading id."""
            if reading.reading_id == reading_id:
                shared[0] = i
                return True

            return False

        self.engine.scan_storage(self.storage_type, _id_searcher)
        found_offset = shared[0]

        if found_offset is None:
            raise UnresolvedIdentifierError("Cannot find reading ID '%d' in storage area '%s'" % (reading_id, self.storage_type))

        return found_offset

    def _verify_offset(self, offset):
        storage_count, streaming_count = self.engine.count()

        if self.storage_type == u'streaming':
            count = streaming_count
        else:
            count = storage_count

        if offset >= count:
            raise UnresolvedIdentifierError("Invalid offset that is larger than the number of valid readings", count=count, offset=offset)
        self.offset = offset

    def peek(self):
        """Peek at the oldest reading in this virtual stream."""

        if self._count == 0:
            raise StreamEmptyError("Peek called on buffered stream walker without any data", selector=self.selector)

        offset = self.offset

        while True:
            curr = self.engine.get(self.storage_type, offset)
            offset += 1

            stream = DataStream.FromEncoded(curr.stream)
            if self.matches(stream):
                return curr

    def skip_all(self):
        """Skip all readings in this walker."""

        storage, streaming = self.engine.count()

        if self.selector.output:
            self.offset = streaming
        else:
            self.offset = storage

        self._count = 0

    def notify_added(self, stream):
        """Notify that a new reading has been added.

        Args:
            stream (DataStream): The stream that had new data
        """

        if not self.matches(stream):
            return

        self._count += 1

    def notify_rollover(self, stream):
        """Notify that a reading in the given stream was overwritten.

        Args:
            stream (DataStream): The stream that had overwritten data.
        """

        self.offset -= 1

        if not self.matches(stream):
            return

        if self._count == 0:
            raise InternalError("BufferedStreamWalker out of sync with storage engine, count was wrong.")

        self._count -= 1


class VirtualStreamWalker(StreamWalker):
    """A stream walker that just walks over virtual streams.

    Args:
        selector (DataStreamSelector): The selector for the stream(s) that we
            are walking.
    """

    def __init__(self, selector):
        super(VirtualStreamWalker, self).__init__(selector)

        if selector.match_id is None:
            raise ArgumentError("You cannot create a stream walker with a wildcard virtual stream")

        self.reading = None

    def dump(self):
        """Serialize the state of this stream walker.

        Returns:
            dict: The serialized state.
        """

        reading = self.reading
        if reading is not None:
            reading = reading.asdict()

        return {
            u'selector': str(self.selector),
            u'reading': reading
        }

    def restore(self, state):
        """Restore the contents of this virtual stream walker.

        Args:
            state (dict): The previously serialized state.

        Raises:
            ArgumentError: If the serialized state does not have
                a matching selector.
        """

        reading = state.get(u'reading')
        if reading is not None:
            reading = IOTileReading.FromDict(reading)

        selector = DataStreamSelector.FromString(state.get(u'selector'))
        if self.selector != selector:
            raise ArgumentError("Attempted to restore a VirtualStreamWalker with a different selector",
                                selector=self.selector, serialized_data=state)

        self.reading = reading

    def push(self, stream, value):
        """Update this stream walker with a new responsive reading.

        Virtual stream walkers keep at most one reading so this function
        just overwrites whatever was previously stored.
        """

        if not self.matches(stream):
            raise ArgumentError("Attempting to push reading to stream walker that does not match", selector=self.selector, stream=stream)

        self.reading = value

    def iter(self):
        """Iterate over the readings that are responsive to this stream walker."""

        if self.reading is not None:
            yield self.reading

    def count(self):
        if self.selector.match_type == DataStream.ConstantType:
            return 0xFFFFFFFF

        if self.reading is None:
            return 0

        return 1

    def pop(self):
        """Pop a reading off of this virtual stream and return it."""

        if self.reading is None:
            raise StreamEmptyError("Pop called on virtual stream walker without any data", selector=self.selector)

        reading = self.reading

        # If we're not a constant stream, we just exhausted ourselves
        if self.selector.match_type != DataStream.ConstantType:
            self.reading = None

        return reading

    def peek(self):
        """Peek at the oldest reading in this virtual stream."""

        if self.reading is None:
            raise StreamEmptyError("Pop called on virtual stream walker without any data", selector=self.selector)

        return self.reading

    def skip_all(self):
        """Skip all readings in this walker."""

        # We can't skip a constant stream
        if self.selector.match_type == DataStream.ConstantType:
            return

        self.reading = None


class CounterStreamWalker(StreamWalker):
    """A stream walker that walks across a counter stream.

    Counter streams only store one reading but keep an accurate
    count of how many times they've been pushed and popped and
    return the last value pushed for each each pop until they're
    empty.

    Args:
        selector (DataStreamSelector): The selector for the stream(s) that
            we're supposed to walk.
    """

    def __init__(self, selector):
        super(CounterStreamWalker, self).__init__(selector)

        if selector.match_id is None:
            raise ArgumentError("You cannot create a stream walker with a wildcard virtual stream")

        self.reading = None
        self._count = 0

    def dump(self):
        """Serialize the state of this stream walker.

        Returns:
            dict: The serialized state.
        """

        reading = self.reading
        if reading is not None:
            reading = reading.asdict()

        return {
            u'selector': str(self.selector),
            u'reading': reading,
            u'count': self._count
        }

    def restore(self, state):
        """Restore the contents of this counter stream walker.

        Args:
            state (dict): The previously serialized state.

        Raises:
            ArgumentError: If the serialized state does not have
                a matching selector.
        """

        reading = state.get(u'reading')
        if reading is not None:
            reading = IOTileReading.FromDict(reading)

        selector = DataStreamSelector.FromString(state.get(u'selector'))
        if self.selector != selector:
            raise ArgumentError("Attempted to restore a CounterStreamWalker with a different selector",
                                selector=self.selector, serialized_data=state)

        count = state.get(u'count')

        self.reading = reading
        self._count = count

    def push(self, stream, value):
        """Update this stream walker with a new responsive reading.

        Args:
            stream (DataStream): The stream that we're pushing
            value (IOTileReading): The reading tha we're pushing
        """

        if not self.matches(stream):
            raise ArgumentError("Attempting to push reading to stream walker that does not match", selector=self.selector, stream=stream)

        self.reading = value
        self._count += 1

    def iter(self):
        """Iterate over the readings that are responsive to this stream walker."""

        for _i in range(0, self._count):
            yield self.reading

    def count(self):
        """Count how many readings are available in this walker."""

        return self._count

    def pop(self):
        """Pop a reading off of this virtual stream and return it."""

        if self._count == 0:
            raise StreamEmptyError("Pop called on virtual stream walker without any data", selector=self.selector)

        self._count = self._count - 1
        return self.reading

    def peek(self):
        """Peek at the oldest reading in this virtual stream."""

        if self.reading is None:
            raise StreamEmptyError("peek called on virtual stream walker without any data", selector=self.selector)

        return self.reading

    def skip_all(self):
        """Skip all readings in this walker."""

        self._count = 0


class InvalidStreamWalker(StreamWalker):
    """A stream walker that is not connected to anything.

    These stream walkers are cannot be used to hold any data and always
    have a count() of 0 but they are useful for unconnected inputs on
    sensor graph node.  The only functions that work are:

        skip_all
        count

    Args:
        selector (DataStreamSelector): The selector for the stream(s) that
            we're supposed to walk.
    """

    def __init__(self, selector):
        super(InvalidStreamWalker, self).__init__(selector)

    def dump(self):
        """Serialize the state of this stream walker.

        Returns:
            dict: The serialized state.
        """

        return {
            u'selector': str(self.selector),
            u'type': u"invalid"
        }

    def restore(self, state):
        """Restore the contents of this virtual stream walker.

        Args:
            state (dict): The previously serialized state.

        Raises:
            ArgumentError: If the serialized state does not have
                a matching selector.
        """


        selector = DataStreamSelector.FromString(state.get(u'selector'))
        if self.selector != selector:
            raise ArgumentError("Attempted to restore an InvalidStreamWalker with a different selector",
                                selector=self.selector, serialized_data=state)

        if state.get(u'type') != u'invalid':
            raise ArgumentError("Invalid serialized state for InvalidStreamWalker", serialized_data=state)

    def matches(self, stream):
        """Check if a stream matches this walker.

        Args:
            stream (DataStream): The stream to check

        Returns:
            bool: Whether the walker matches the stream
        """

        return False

    def push(self, stream, value):
        """Update this stream walker with a new responsive reading.

        Args:
            stream (DataStream): The stream that we're pushing
            value (IOTileReading): The reading tha we're pushing
        """

        raise ArgumentError("Attempting to push reading to an invalid stream walker that cannot hold data", selector=self.selector, stream=stream)


    def iter(self):
        """Iterate over the readings that are responsive to this stream walker."""

        return []

    def count(self):
        """Count how many readings are available in this walker."""

        return 0

    def pop(self):
        """Pop a reading off of this virtual stream and return it."""

        raise StreamEmptyError("Pop called on an invalid stream walker")

    def peek(self):
        """Peek at the oldest reading in this virtual stream."""

        raise StreamEmptyError("peek called on an invalid stream walker")

    def skip_all(self):
        """Skip all readings in this walker."""

        pass



import pytest

from iotile.core.exceptions import ArgumentError
from iotile.sg.model import DeviceModel
from iotile.sg.sensor_log import SensorLog
from iotile.sg.exceptions import StorageFullError, UnresolvedIdentifierError
from iotile.sg.engine import InMemoryStorageEngine
from iotile.sg import DataStreamSelector, DataStream, StreamEmptyError
from iotile.core.hw.reports import IOTileReading


def test_counter_walker():
    """Make sure counter walkers work correctly."""

    model = DeviceModel()
    log = SensorLog(model=model)

    walk = log.create_walker(DataStreamSelector.FromString('counter 1'))

    stream = DataStream.FromString('counter 1')
    reading = IOTileReading(0, 1, 1)

    assert walk.count() == 0
    log.push(stream, IOTileReading(0, 1, 1))
    assert walk.count() == 1
    assert walk.peek().value == 1

    log.push(stream, IOTileReading(0, 1, 3))

    assert walk.count() == 2
    assert walk.peek().value == 3

    val = walk.pop()
    assert walk.count() == 1
    assert val.value == 3

    val = walk.pop()
    assert walk.count() == 0
    assert val.value == 3

    with pytest.raises(StreamEmptyError):
        walk.pop()


def test_unbuffered_walker():
    """Make sure unbuffered walkers hold only 1 reading."""

    model = DeviceModel()
    log = SensorLog(model=model)

    walk = log.create_walker(DataStreamSelector.FromString('unbuffered 1'))
    stream = DataStream.FromString('unbuffered 1')

    assert walk.count() == 0
    log.push(stream, IOTileReading(0, 1, 1))
    assert walk.count() == 1
    assert walk.peek().value == 1

    log.push(stream, IOTileReading(0, 1, 3))
    assert walk.count() == 1
    assert walk.peek().value == 3

    val = walk.pop()
    assert walk.count() == 0
    assert val.value == 3

    with pytest.raises(StreamEmptyError):
        walk.pop()


def test_constant_walker():
    """Make sure constant walkers can be read any number of times."""

    model = DeviceModel()
    log = SensorLog(model=model)

    walk = log.create_walker(DataStreamSelector.FromString('constant 1'))
    stream = DataStream.FromString('constant 1')
    reading = IOTileReading(0, 1, 1)

    log.push(stream, reading)
    assert walk.count() == 0xFFFFFFFF

    log.push(stream, reading)
    assert walk.count() == 0xFFFFFFFF

    val = walk.pop()
    assert walk.count() == 0xFFFFFFFF

    val = walk.pop()
    assert walk.count() == 0xFFFFFFFF


def test_storage_walker():
    """Make sure the storage walker works."""

    model = DeviceModel()
    log = SensorLog(model=model)


    walk = log.create_walker(DataStreamSelector.FromString('buffered 1'))
    stream = DataStream.FromString('buffered 1')

    storage_size = model.get('max_storage_buffer')
    erase_size = model.get('buffer_erase_size')

    for i in range(0, storage_size):
        reading = IOTileReading(0, stream.encode(), i)
        log.push(stream, reading)

        assert walk.count() == (i + 1)

    # Make sure the overwrite happens correctly
    old_count = walk.count()
    reading = IOTileReading(0, stream.encode(), storage_size)
    log.push(stream, reading)
    assert walk.count() == (old_count - erase_size + 1)


def test_walker_at_beginning():
    """Make sure we can start a walker at the beginning of a stream."""

    model = DeviceModel()
    log = SensorLog(model=model)

    stream = DataStream.FromString('buffered 1')
    reading = IOTileReading(stream.encode(), 0, 1)
    log.push(stream, reading)
    log.push(stream, reading)
    log.push(DataStream.FromString('buffered 2'), reading)

    walk = log.create_walker(DataStreamSelector.FromString('buffered 1'), skip_all=False)
    assert walk.offset == 0
    assert walk.count() == 2


def test_storage_streaming_walkers():
    """Make sure the storage and streaming walkers work simultaneously."""

    model = DeviceModel()
    log = SensorLog(model=model)


    storage_walk = log.create_walker(DataStreamSelector.FromString('buffered 1'))
    output_walk = log.create_walker(DataStreamSelector.FromString('output 1'))
    storage1 = DataStream.FromString('buffered 1')
    storage2 = DataStream.FromString('buffered 2')
    output1 = DataStream.FromString('output 1')
    output2 = DataStream.FromString('output 2')

    assert storage_walk.offset == 0
    assert output_walk.offset == 0

    for i in range(0, 5000):
        reading = IOTileReading(0, 0, i)
        log.push(storage1, reading)
        log.push(storage2, reading)
        log.push(output1, reading)
        log.push(output2, reading)

        assert storage_walk.count() == (i + 1)
        assert output_walk.count() == (i + 1)
        assert storage_walk.offset == 0
        assert output_walk.offset == 0

    for i in range(0, 5000):
        reading1 = storage_walk.pop()
        reading2 = output_walk.pop()

        assert reading1.value == i
        assert reading1.stream == storage1.encode()
        assert reading2.value == i
        assert reading2.stream == output1.encode()
        assert storage_walk.offset == (2 * i) + 1
        assert output_walk.offset == (2* i) + 1

    log.clear()

    assert storage_walk.count() == 0
    assert storage_walk.offset == 0

    assert output_walk.count() == 0
    assert output_walk.offset == 0


def test_storage_scan():
    """Make sure scan_storage works."""

    model = DeviceModel()

    engine = InMemoryStorageEngine(model)
    log = SensorLog(engine, model=model)

    storage1 = DataStream.FromString('buffered 1')
    output1 = DataStream.FromString('output 1')

    for i in range(0, 10):
        reading = IOTileReading(0, 0, i)
        log.push(storage1, reading)

    for i in range(7, 14):
        reading = IOTileReading(0, 0, i)
        log.push(output1, reading)

    shared = [None, 0]

    def _max_counter(i, reading):
        if reading.value > shared[1]:
            shared[0] = i
            shared[1] = reading.value

    # Make sure scanning storage works
    shared = [None, 0]
    count = engine.scan_storage('storage', _max_counter)
    assert count == 10
    assert shared == [9, 9]

    shared = [None, 0]
    count = engine.scan_storage('storage', _max_counter, start=5, stop=9)
    assert count == 5
    assert shared == [9, 9]

    shared = [None, 0]
    count = engine.scan_storage('storage', _max_counter, start=5, stop=8)
    assert count == 4
    assert shared == [8, 8]

    # Make sure scanning steaming works
    shared = [None, 0]
    count = engine.scan_storage('streaming', _max_counter)
    assert count == 7
    assert shared == [6, 13]

    shared = [None, 0]
    count = engine.scan_storage('streaming', _max_counter, start=2, stop=6)
    assert count == 5
    assert shared == [6, 13]

    shared = [None, 0]
    count = engine.scan_storage('streaming', _max_counter, start=2, stop=5)
    assert count == 4
    assert shared == [5, 12]

    # Make sure errors are thrown correctly
    with pytest.raises(ArgumentError):
        engine.scan_storage('other_name', _max_counter)

    with pytest.raises(ArgumentError):
        engine.scan_storage('streaming', _max_counter, stop=7)


def test_seek_walker():
    """Make sure we can seek a walker can count with an offset."""

    model = DeviceModel()
    log = SensorLog(model=model)

    stream = DataStream.FromString('buffered 1')
    reading = IOTileReading(stream.encode(), 0, 1)
    log.push(stream, reading)
    log.push(stream, reading)
    log.push(DataStream.FromString('buffered 2'), reading)
    log.push(stream, reading)

    walk = log.create_walker(DataStreamSelector.FromString('buffered 1'), skip_all=False)
    assert walk.offset == 0
    assert walk.count() == 3

    walk.seek(1)
    assert walk.offset == 1
    assert walk.count() == 2

    # Make sure we can seek to a position corresponding to another stream
    walk.seek(2)
    assert walk.offset == 2
    assert walk.count() == 1

    # Make sure we can find a reading by reading ID
    walk = log.create_walker(DataStreamSelector.FromString('output 1'), skip_all=False)

    output1 = DataStream.FromString('output 1')
    output2 = DataStream.FromString('output 2')
    log.push(output1, IOTileReading(0, 0, 1, reading_id=1))
    log.push(output1, IOTileReading(0, 0, 1, reading_id=2))
    log.push(output2, IOTileReading(0, 0, 1, reading_id=3))
    log.push(output1, IOTileReading(0, 0, 1, reading_id=4))

    exact = walk.seek(2, target='id')
    assert exact is True
    assert walk.count() == 2
    assert walk.offset == 1

    exact = walk.seek(3, target='id')
    assert exact is False
    assert walk.count() == 1
    assert walk.offset == 2

    # Verify exceptions thrown by seek()
    with pytest.raises(UnresolvedIdentifierError):
        walk.seek(5, target='id')

    with pytest.raises(UnresolvedIdentifierError):
        walk.seek(5, target=u'id')

    with pytest.raises(UnresolvedIdentifierError):
        walk.seek(5, target='offset')

    with pytest.raises(ArgumentError):
        walk.seek(2, target="unsupported")

def test_fill_stop():
    """Make sure we can configure SensorLog into fill-stop mode."""

    model = DeviceModel()
    log = SensorLog(model=model)

    storage = DataStream.FromString('buffered 1')
    output = DataStream.FromString('output 1')

    reading = IOTileReading(0, 0, 1)

    # Neither fill-stop
    for _i in range(0, 50000):
        log.push(storage, reading)

    for _i in range(0, 50000):
        log.push(output, reading)

    log.clear()
    log.set_rollover('storage', False)

    with pytest.raises(StorageFullError):
        for _i in range(0, 50000):
            log.push(storage, reading)

    for _i in range(0, 50000):
        log.push(output, reading)

    assert log.count() == (16128, 48720)

    log.clear()
    log.set_rollover('streaming', False)

    with pytest.raises(StorageFullError):
        for _i in range(0, 50000):
            log.push(storage, reading)

    with pytest.raises(StorageFullError):
        for _i in range(0, 50000):
            log.push(output, reading)

    assert log.count() == (16128, 48896)


def test_dump_restore():
    """Make sure we can properly dump and restore a SensorLog."""

    model = DeviceModel()
    log = SensorLog(model=model)

    storage = DataStream.FromString('buffered 1')
    output = DataStream.FromString('output 1')

    reading = IOTileReading(0, 0, 1)

    for _i in range(0, 25):
        log.push(storage, reading)

    for _i in range(0, 20):
        log.push(output, reading)

    out1 = log.create_walker(DataStreamSelector.FromString('output 1'), skip_all=False)
    store1 = log.create_walker(DataStreamSelector.FromString('buffered 1'), skip_all=False)
    count1 = log.create_walker(DataStreamSelector.FromString('counter 1'))
    const1 = log.create_walker(DataStreamSelector.FromString('constant 1'))
    unbuf1 = log.create_walker(DataStreamSelector.FromString('unbuffered 1'))

    log.push(DataStream.FromString('counter 1'), reading)
    log.push(DataStream.FromString('counter 1'), reading)
    log.push(DataStream.FromString('constant 1'), reading)
    log.push(DataStream.FromString('unbuffered 1'), reading)

    state = log.dump()

    log.clear()
    log.destroy_all_walkers()

    out1 = log.create_walker(DataStreamSelector.FromString('output 1'), skip_all=False)
    store1 = log.create_walker(DataStreamSelector.FromString('buffered 1'), skip_all=False)
    count1 = log.create_walker(DataStreamSelector.FromString('counter 1'))
    const1 = log.create_walker(DataStreamSelector.FromString('constant 1'))
    unbuf1 = log.create_walker(DataStreamSelector.FromString('unbuffered 1'))

    log.restore(state)

    assert store1.count() == 25
    assert out1.count() == 20
    assert count1.count() == 2
    assert const1.count() == 0xFFFFFFFF
    assert unbuf1.count() == 1

    # Test permissive and non-permissive restores
    _unbuf2 = log.create_walker(DataStreamSelector.FromString('unbuffered 2'))

    with pytest.raises(ArgumentError):
        log.restore(state)

    log.clear()
    log.restore(state, permissive=True)
    assert store1.count() == 25
    assert out1.count() == 20
    assert count1.count() == 2
    assert const1.count() == 0xFFFFFFFF
    assert unbuf1.count() == 1

    # Test restoring a stream walker
    log.clear()
    log.destroy_all_walkers()
    walk = log.create_walker(DataStreamSelector.FromString(str(storage)))

    for _i in range(0, 25):
        log.push(storage, reading)

    assert walk.count() == 25
    dump = walk.dump()
    log.destroy_all_walkers()
    walk2 = log.restore_walker(dump)
    assert walk2.count() == 25



"""An in memory storage engine for sensor graph."""

from builtins import str, range
from iotile.core.exceptions import ArgumentError
from iotile.core.hw.reports import IOTileReading
from iotile.sg import DataStream
from iotile.sg.exceptions import StorageFullError, StreamEmptyError


class InMemoryStorageEngine(object):
    """A simple in memory storage engine for sensor graph.

    Args:
        model (DeviceModel): A model for the device type that we are
            emulating so that we can constrain our total memory
            size appropriately to get the same behavior that would
            be seen on an actual device.
    """

    def __init__(self, model):
        self.model = model
        self.storage_length = model.get(u'max_storage_buffer')
        self.streaming_length = model.get(u'max_streaming_buffer')
        self.streaming_data = []
        self.storage_data = []

    def dump(self):
        """Serialize the state of this InMemoryStorageEngine to a dict.

        Returns:
            dict: The serialized data.
        """

        return {
            u'storage_data': [x.asdict() for x in self.storage_data],
            u'streaming_data': [x.asdict() for x in self.streaming_data]
        }

    def restore(self, state):
        """Restore the state of this InMemoryStorageEngine from a dict."""

        storage_data = state.get(u'storage_data', [])
        streaming_data = state.get(u'streaming_data', [])

        if len(storage_data) > self.storage_length or len(streaming_data) > self.streaming_length:
            raise ArgumentError("Cannot restore InMemoryStorageEngine, too many readings",
                                storage_size=len(storage_data), storage_max=self.storage_length,
                                streaming_size=len(streaming_data), streaming_max=self.streaming_length)

        self.storage_data = [IOTileReading.FromDict(x) for x in storage_data]
        self.streaming_data = [IOTileReading.FromDict(x) for x in streaming_data]

    def count(self):
        """Count the number of readings.

        Returns:
            (int, int): The number of readings in storage and streaming buffers.
        """

        return (len(self.storage_data), len(self.streaming_data))

    def count_matching(self, selector, offset=0):
        """Count the number of readings matching selector.

        Args:
            selector (DataStreamSelector): The selector that we want to
                count matching readings for.
            offset (int): The starting offset that we should begin counting at.

        Returns:
            int: The number of matching readings.
        """

        if selector.output:
            data = self.streaming_data
        elif selector.buffered:
            data = self.storage_data
        else:
            raise ArgumentError("You can only pass a buffered selector to count_matching", selector=selector)

        count = 0
        for i in range(offset, len(data)):
            reading = data[i]

            stream = DataStream.FromEncoded(reading.stream)
            if selector.matches(stream):
                count += 1

        return count

    def scan_storage(self, area_name, callable, start=0, stop=None):
        """Iterate over streaming or storage areas, calling callable.

        Args:
            area_name (str): Either 'storage' or 'streaming' to indicate which
                storage area to scan.
            callable (callable): A function that will be called as (offset, reading)
                for each reading between start_offset and end_offset (inclusive).  If
                the scan function wants to stop early it can return True.  If it returns
                anything else (including False or None), scanning will continue.
            start (int): Optional offset to start at (included in scan).
            stop (int): Optional offset to end at (included in scan).

        Returns:
            int: The number of entries scanned.
        """

        if area_name == u'storage':
            data = self.storage_data
        elif area_name == u'streaming':
            data = self.streaming_data
        else:
            raise ArgumentError("Unknown area name in scan_storage (%s) should be storage or streaming" % area_name)

        if len(data) == 0:
            return 0

        if stop is None:
            stop = len(data) - 1
        elif stop >= len(data):
            raise ArgumentError("Given stop offset is greater than the highest offset supported", length=len(data), stop_offset=stop)

        scanned = 0
        for i in range(start, stop + 1):
            scanned += 1

            should_break = callable(i, data[i])
            if should_break is True:
                break

        return scanned

    def clear(self):
        """Clear all data from this storage engine."""

        self.storage_data = []
        self.streaming_data = []

    def push(self, value):
        """Store a new value for the given stream.

        Args:
            value (IOTileReading): The value to store.  The stream
                parameter must have the correct value
        """

        stream = DataStream.FromEncoded(value.stream)

        if stream.stream_type == DataStream.OutputType:
            if len(self.streaming_data) == self.streaming_length:
                raise StorageFullError('Streaming buffer full')

            self.streaming_data.append(value)
        else:
            if len(self.storage_data) == self.storage_length:
                raise StorageFullError('Storage buffer full')

            self.storage_data.append(value)

    def get(self, buffer_type, offset):
        """Get a reading from the buffer at offset.

        Offset is specified relative to the start of the data buffer.
        This means that if the buffer rolls over, the offset for a given
        item will appear to change.  Anyone holding an offset outside of this
        engine object will need to be notified when rollovers happen (i.e.
        popn is called so that they can update their offset indices)

        Args:
            buffer_type (str): The buffer to pop from (either u"storage" or u"streaming")
            offset (int): The offset of the reading to get
        """

        if buffer_type == u'streaming':
            chosen_buffer = self.streaming_data
        else:
            chosen_buffer = self.storage_data

        if offset >= len(chosen_buffer):
            raise StreamEmptyError("Invalid index given in get command", requested=offset, stored=len(chosen_buffer), buffer=buffer_type)

        return chosen_buffer[offset]

    def popn(self, buffer_type, count):
        """Remove and return the oldest count values from the named buffer

        Args:
            buffer_type (str): The buffer to pop from (either u"storage" or u"streaming")
            count (int): The number of readings to pop

        Returns:
            list(IOTileReading): The values popped from the buffer
        """

        buffer_type = str(buffer_type)

        if buffer_type == u'streaming':
            chosen_buffer = self.streaming_data
        else:
            chosen_buffer = self.storage_data

        if count > len(chosen_buffer):
            raise StreamEmptyError("Not enough data in buffer for popn command", requested=count, stored=len(chosen_buffer), buffer=buffer_type)

        popped = chosen_buffer[:count]
        remaining = chosen_buffer[count:]

        if buffer_type == u'streaming':
            self.streaming_data = remaining
        else:
            self.storage_data = remaining

        return popped



"""Parsing routines for understanding binary streamer descriptors.

Binary streamer descriptors are the internal binary representation of
a streamer stored in an IOTile device's embedded firmware.
"""

from __future__ import (unicode_literals, absolute_import, print_function)
import struct
from builtins import str
from typedargs.exceptions import ArgumentError
from .slot import SlotIdentifier
from .stream import DataStreamSelector
from .streamer import DataStreamer
from .exceptions import SensorGraphSemanticError
from .parser.language import get_streamer_parser


def parse_binary_descriptor(bindata, sensor_log=None):
    """Convert a binary streamer descriptor into a string descriptor.

    Binary streamer descriptors are 20-byte binary structures that encode all
    information needed to create a streamer.  They are used to communicate
    that information to an embedded device in an efficent format.  This
    function exists to turn such a compressed streamer description back into
    an understandable string.

    Args:
        bindata (bytes): The binary streamer descriptor that we want to
            understand.
        sensor_log (SensorLog): Optional sensor_log to add this streamer to
            a an underlying data store.

    Returns:
        DataStreamer: A DataStreamer object representing the streamer.

        You can get a useful human readable string by calling str() on the
        return value.
    """

    if len(bindata) != 14:
        raise ArgumentError("Invalid length of binary data in streamer descriptor", length=len(bindata), expected=14, data=bindata)

    dest_tile, stream_id, trigger, format_code, type_code = struct.unpack("<8sHBBBx", bindata)

    dest_id = SlotIdentifier.FromEncoded(dest_tile)
    selector = DataStreamSelector.FromEncoded(stream_id)

    format_name = DataStreamer.KnownFormatCodes.get(format_code)
    type_name = DataStreamer.KnownTypeCodes.get(type_code)

    if format_name is None:
        raise ArgumentError("Unknown format code", code=format_code, known_code=DataStreamer.KnownFormatCodes)
    if type_name is None:
        raise ArgumentError("Unknown type code", code=type_code, known_codes=DataStreamer.KnownTypeCodes)

    with_other = None
    if trigger & (1 << 7):
        auto = False
        with_other = trigger & ((1 << 7) - 1)
    elif trigger == 0:
        auto = False
    elif trigger == 1:
        auto = True
    else:
        raise ArgumentError("Unknown trigger type for streamer", trigger_code=trigger)

    return DataStreamer(selector, dest_id, format_name, auto, type_name, with_other=with_other, sensor_log=sensor_log)


def create_binary_descriptor(streamer):
    """Create a packed binary descriptor of a DataStreamer object.

    Args:
        streamer (DataStreamer): The streamer to create a packed descriptor for

    Returns:
        bytes: A packed 14-byte streamer descriptor.
    """

    trigger = 0
    if streamer.automatic:
        trigger = 1
    elif streamer.with_other is not None:
        trigger = (1 << 7) | streamer.with_other

    return struct.pack("<8sHBBBx", streamer.dest.encode(), streamer.selector.encode(), trigger, streamer.KnownFormats[streamer.format], streamer.KnownTypes[streamer.report_type])


def parse_string_descriptor(string_desc):
    """Parse a string descriptor of a streamer into a DataStreamer object.

    Args:
        string_desc (str): The string descriptor that we wish to parse.

    Returns:
        DataStreamer: A DataStreamer object representing the streamer.
    """

    if not isinstance(string_desc, str):
        string_desc = str(string_desc)

    if not string_desc.endswith(';'):
        string_desc += ';'

    parsed = get_streamer_parser().parseString(string_desc)[0]

    realtime = 'realtime' in parsed
    broadcast = 'broadcast' in parsed
    encrypted = 'security' in parsed and parsed['security'] == 'encrypted'
    signed = 'security' in parsed and parsed['security'] == 'signed'
    auto = 'manual' not in parsed

    with_other = None
    if 'with_other' in parsed:
        with_other = parsed['with_other']
        auto = False

    dest = SlotIdentifier.FromString('controller')
    if 'explicit_tile' in parsed:
        dest = parsed['explicit_tile']

    selector = parsed['selector']

    # Make sure all of the combination are valid
    if realtime and (encrypted or signed):
        raise SensorGraphSemanticError("Realtime streamers cannot be either signed or encrypted")

    if broadcast and (encrypted or signed):
        raise SensorGraphSemanticError("Broadcast streamers cannot be either signed or encrypted")

    report_type = 'broadcast' if broadcast else 'telegram'
    dest = dest
    selector = selector

    if realtime or broadcast:
        report_format = u'individual'
    elif signed:
        report_format = u'signedlist_userkey'
    elif encrypted:
        raise SensorGraphSemanticError("Encrypted streamers are not yet supported")
    else:
        report_format = u'hashedlist'

    return DataStreamer(selector, dest, report_format, auto, report_type=report_type, with_other=with_other)



"""Tests of DataStreamer functionality."""

from iotile.sg.streamer_descriptor import parse_binary_descriptor, parse_string_descriptor, create_binary_descriptor


def test_descriptors():
    """Make sure we can parse string and binary descriptors."""
    streamer = parse_string_descriptor('manual realtime streamer on unbuffered 1 to controller')
    assert str(streamer) == 'manual realtime streamer on unbuffered 1'

    bin_desc = create_binary_descriptor(streamer)
    streamer2 = parse_binary_descriptor(bin_desc)

    assert streamer2 == streamer

    streamer = parse_string_descriptor('realtime streamer on unbuffered 1 to slot 5 with streamer 0')
    assert str(streamer) == 'manual realtime streamer on unbuffered 1 to slot 5 with streamer 0'

    streamer = parse_string_descriptor(u'realtime streamer on unbuffered 1 to slot 5 with streamer 0')
    assert str(streamer) == 'manual realtime streamer on unbuffered 1 to slot 5 with streamer 0'

    bin_desc = create_binary_descriptor(streamer)
    streamer2 = parse_binary_descriptor(bin_desc)

    assert streamer2 == streamer



"""Configuration object describing a streamer."""

from __future__ import unicode_literals, absolute_import, print_function
from collections import namedtuple
from future.utils import viewitems, python_2_unicode_compatible
from iotile.core.hw.reports import IndividualReadingReport, BroadcastReport, SignedListReport
from iotile.core.exceptions import ArgumentError, InternalError
from iotile.sg.exceptions import StreamEmptyError

StreamerReport = namedtuple("StreamerReport", ['report', 'num_readings', 'highest_id'])

@python_2_unicode_compatible
class DataStreamer(object):
    """In charge of streaming data remotely from a device via a communication tile.

    DataStreamers are created to listen for data from a stream, package it using some
    packaging scheme and then attempt to send it remotely when triggered (possibly
    automatically by the presence of data).  They can be configured to keep trying to
    send data until it is accepted by the other side.

    Args:
        selector (DataStreamSelector): The selector that chooses what readings will be
            streamed.
        dest_tile (SlotIdentifier): The destination tile that we will stream data to
        report_format (str): A string constant defined in DataStreamer.KnownFormats
        automatic (bool): Whether the streamer should trigger whenever there is data
            or wait until it is manually triggered
        report_type (str): A string constant defined in DataStreamer.KnownTypes
        with_other (int): The index of another streamer that this streamer should
            use to know when to trigger.  Defaults to None in which case the streamer
            triggers on its own.  The combination of automatic and with_other cannot
            be specified.
        sensor_log (SensorLog): Actually create a StreamWalker to go along with this
            streamer so that we can check if it's triggered.
    """

    KnownTypes = {u'broadcast': 1, u'telegram': 1 << 1, u'synchronous': 1 << 2}
    KnownTypeCodes = {y: x for x, y in viewitems(KnownTypes)}
    KnownFormats = {u'individual': 0, u'hashedlist': 1, u'signedlist_userkey': 2, u'signedlist_devicekey': 3}
    KnownFormatCodes = {y: x for x, y in viewitems(KnownFormats)}

    def __init__(self, selector, dest_tile, report_format, automatic, report_type=u'telegram', with_other=None, sensor_log=None):
        report_format = str(report_format)
        report_type = str(report_type)

        if report_format not in DataStreamer.KnownFormats:
            raise ArgumentError("Unknown report format in DataStreamer constructor", report_format=report_format, known_formats=DataStreamer.KnownFormats.keys())

        if report_type not in DataStreamer.KnownTypes:
            raise ArgumentError("Unknown report type in DataStreamer constructor", report_type=report_type, known_types=DataStreamer.KnownTypes.keys())

        self.selector = selector
        self.dest = dest_tile
        self.format = report_format
        self.automatic = automatic
        self.report_type = report_type
        self.with_other = with_other
        self.walker = None
        self.index = None
        self._sensor_log = None

        if sensor_log is not None:
            self.link_to_storage(sensor_log)

    def link_to_storage(self, sensor_log):
        """Attach this DataStreamer to an underlying SensorLog.

        Calling this method is required if you want to use this DataStreamer
        to generate reports from the underlying data in the SensorLog.

        You can call it multiple times and it will unlink itself from any
        previous SensorLog each time.

        Args:
            sensor_log (SensorLog): Actually create a StreamWalker to go along with this
                streamer so that we can check if it's triggered.
        """

        if self.walker is not None:
            self._sensor_log.destroy_walker(self.walker)
            self.walker = None

        self.walker = sensor_log.create_walker(self.selector)
        self._sensor_log = sensor_log

    def has_data(self):
        """Check whether there is any data in this streamer.

        Returns:
            bool: Whether there is any available data.
        """

        if self.walker is None:
            raise InternalError("You can only check if a streamer is triggered if you create it with a SensorLog")

        return self.walker.count() > 0

    def triggered(self, manual=False):
        """Check if this streamer should generate a report.

        Streamers can be triggered automatically whenever they have data
        or they can be triggered manually. This method returns True if the
        streamer is currented triggered.

        A streamer is triggered if it:
          - (has data AND is automatic) OR
          - (has data AND is manually triggered)

        Args:
            manual (bool): Indicate that the streamer has been manually triggered.

        Returns:
            bool: Whether the streamer can generate a report right now.
        """

        if self.walker is None:
            raise InternalError("You can only check if a streamer is triggered if you create it with a SensorLog")

        if not self.automatic and not manual:
            return False

        return self.has_data()

    def requires_id(self):
        """Whether this streamer produces reports that require a report id.

        Returns:
            bool
        """

        return self.format != u'individual'

    def requires_signing(self):
        """Whether this streamer produces reports that require a valid auth_chain for signing.

        Returns:
            bool
        """

        return self.format in (u'signedlist_userkey', u'signedlist_devicekey')

    def build_report(self, device_id, max_size=None, device_uptime=0, report_id=None, auth_chain=None):
        """Build a report with all of the readings in this streamer.

        This method will produce an IOTileReport subclass and, if necessary,
        sign it using the passed authentication chain.

        Args:
            device_id (int): The UUID of the device to generate a report for.
            max_size (int): Optional maximum number of bytes that the report can be
            device_uptime (int): The device's uptime to use as the sent timestamp of the report
            report_id (int): The report id to use if the report type require serialization.
            auth_chain (AuthChain): An auth chain class to use to sign the report if the report
                type requires signing.

        Returns:
            StreamerReport: The report, its highest id and the number of readings in it.

            The highest reading id and number of readings are returned
            separately from the report itself because, depending on the format
            of the report (such as whether it is encrypted or does not contain
            reading ids), these details may not be recoverable from the report
            itself.

        Raises:
            InternalError: If there was no SensorLog passed when this streamer was created.
            StreamEmptyError: If there is no data to generate a report from.  This can only happen
                if a call to triggered() returned False.
            ArgumentError: If the report requires additional metadata that was not passed like a
                signing key or report_id.
        """

        if self.walker is None or self.index is None:
            raise InternalError("You can only build a report with a DataStreamer if you create it with a SensorLog and a streamer index")

        if self.requires_signing() and auth_chain is None:
            raise ArgumentError("You must pass an auth chain to sign this report.")

        if self.requires_id() and report_id is None:
            raise ArgumentError("You must pass a report_id to serialize this report")

        if self.format == 'individual':
            reading = self.walker.pop()

            highest_id = reading.reading_id

            if self.report_type == 'telegram':
                return StreamerReport(IndividualReadingReport.FromReadings(device_id, [reading]), 1, highest_id)
            elif self.report_type == 'broadcast':
                return StreamerReport(BroadcastReport.FromReadings(device_id, [reading], device_uptime), 1, highest_id)
        elif self.format == 'hashedlist':
            max_readings = (max_size - 20 - 24) // 16
            if max_readings <= 0:
                raise InternalError("max_size is too small to hold even a single reading", max_size=max_size)

            readings = []
            highest_id = 0
            try:
                while len(readings) < max_readings:
                    reading = self.walker.pop()
                    readings.append(reading)
                    if reading.reading_id > highest_id:
                        highest_id = reading.reading_id
            except StreamEmptyError:
                if len(readings) == 0:
                    raise

            return StreamerReport(SignedListReport.FromReadings(device_id, readings, report_id=report_id, selector=self.selector.encode(),
                                                                streamer=self.index, sent_timestamp=device_uptime), len(readings), highest_id)

        raise InternalError("Streamer report format or type is not supported currently", report_format=self.format, report_type=self.report_type)

    def __str__(self):
        manual = "manual " if not self.automatic else ""
        if self.report_type == 'broadcast':
            type_string = 'broadcast '
        elif self.format == 'individual':
            type_string = 'realtime '
        else:
            type_string = ""

        security = ""
        if self.format == 'signedlist_userkey':
            security = "signed "

        to_slot = ""
        if not self.dest.controller:
            to_slot = " to " + str(self.dest)

        #FIXME: Make sure we generate the right code for streamer type

        with_statement = ""
        if self.with_other is not None:
            with_statement = " with streamer %d" % self.with_other

        template = "{manual}{security}{report_type}streamer on {selector}{to_slot}{with_other}"
        return template.format(manual=manual, security=security, report_type=type_string, selector=self.selector,
                               with_other=with_statement, to_slot=to_slot)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if not isinstance(other, DataStreamer):
            return NotImplemented

        return (self.selector == other.selector and self.dest == other.dest and
                self.automatic == other.automatic and self.report_type == other.report_type and
                self.with_other == other.with_other)

    def __ne__(self, other):
        if not isinstance(other, DataStreamer):
            return NotImplemented

        return not self == other



"""A centralized data management structure that allows multiple named FIFO streams.

SensorGraph is organized around data streams, which are named FIFOs that hold
readings.  For the purposes of embedding SensorGraph into a low resource setting
you can limit the number of readings stored in any given FIFO as needed to put
a hard cap on storage requirements.
"""

import copy
from future.utils import viewitems
from iotile.sg.model import DeviceModel
from iotile.core.exceptions import ArgumentError
from iotile.core.hw.reports import IOTileReading
from .engine import InMemoryStorageEngine
from .stream import DataStream, DataStreamSelector
from .walker import VirtualStreamWalker, CounterStreamWalker, BufferedStreamWalker
from .exceptions import StreamEmptyError, StorageFullError, UnresolvedIdentifierError


class SensorLog(object):
    """A storage engine holding multiple named FIFOs.

    Normally a SensorLog is used in ring-buffer mode which means that old
    readings are automatically overwritten as needed when new data is saved.

    However, you can configure it into fill-stop mode by using:
    set_rollover("streaming"|"storage", True|False)

    By default rollover is set to True for both streaming and storage and can
    be controlled individually for each one.

    Readings that are committed to persistent storage streams are assigned
    unique identifiers.  If you have an external system for setting reading
    ids, you can send them to push() already assigned with correct reading
    ids.  If you would prefer to have SensorLog assign the IDs for you, you
    can do so by setting the id_assigner parameter to a callable that will be
    called with the stream and reading and should return an integer.

    Args:
        engine (StorageEngine): The engine used for storing persistent data
            generated by this sensor graph. This can either be a simple, in
            memory data store, or a more complicated persistent storage setup
            depending on needs.  If not specified, a temporary in memory
            engine is used.
        model (DeviceModel): An optional device model specifying the
            constraints of the device that we are emulating.  If not
            specified, no specific constraints are imposed on the sensor log.
        id_assigner (callable): Optional callable function that will be used
            to assign the reading_id to each IOTileReading that is committed
            to persistent storage.  You can also set this later by using the
            id_assigner property.  If you do not specify this parameter or
            set it later, then it will be assumed that the readings sent
            to push() already have correct reading_id numbers.
    """

    def __init__(self, engine=None, model=None, id_assigner=None):
        self._monitors = {}
        self._last_values = {}
        self._virtual_walkers = []
        self._queue_walkers = []

        if model is None:
            model = DeviceModel()

        if engine is None:
            engine = InMemoryStorageEngine(model=model)

        self._engine = engine
        self._model = model
        self._rollover_streaming = True
        self._rollover_storage = True

        self.id_assigner = id_assigner

    def dump(self):
        """Dump the state of this SensorLog.

        The purpose of this method is to be able to restore the same state
        later.  However there are links in the SensorLog for stream walkers.

        So the dump process saves the state of each stream walker and upon
        restore, it looks through the current set of stream walkers and
        restores each one that existed when dump() was called to its state.

        Returns:
            dict: The serialized state of this SensorLog.
        """

        walkers = {}
        walkers.update({str(walker.selector): walker.dump() for walker in self._queue_walkers})
        walkers.update({str(walker.selector): walker.dump() for walker in self._virtual_walkers})

        return {
            u'engine': self._engine.dump(),
            u'rollover_storage': self._rollover_storage,
            u'rollover_streaming': self._rollover_streaming,
            u'last_values': {str(stream): reading.asdict() for stream, reading in viewitems(self._last_values)},
            u'walkers': walkers
        }

    def restore(self, state, permissive=False):
        """Restore a state previously dumped by a call to dump().

        The purpose of this method is to be able to restore a previously
        dumped state.  However there are links in the SensorLog for stream
        walkers.

        So the restore process looks through the current set of stream walkers
        and restores each one that existed when dump() was called to its
        state.  If there are walkers allocated that were not present when
        dump() was called, an exception is raised unless permissive=True,
        in which case they are ignored.

        Args:
            state (dict): The previous state to restore, from a prior call
                to dump().
            permissive (bool): Whether to raise an exception is new stream
                walkers are present that do not have dumped contents().

        Raises:
            ArgumentError: There are new stream walkers present in the current
                SensorLog and permissive==False.
        """

        self._engine.restore(state.get(u'engine'))
        self._last_values = {DataStream.FromString(stream): IOTileReading.FromDict(reading) for
                             stream, reading in viewitems(state.get(u"last_values", {}))}

        self._rollover_storage = state.get(u'rollover_storage', True)
        self._rollover_streaming = state.get(u'rollover_streaming', True)

        old_walkers = {DataStreamSelector.FromString(selector): dump for selector, dump in
                       viewitems(state.get(u"walkers"))}

        for walker in self._virtual_walkers:
            if walker.selector in old_walkers:
                walker.restore(old_walkers[walker.selector])
            elif not permissive:
                raise ArgumentError("Cannot restore SensorLog, walker %s exists in restored log but did not exist before" % str(walker.selector))

        for walker in self._queue_walkers:
            if walker.selector in old_walkers:
                walker.restore(old_walkers[walker.selector])
            elif not permissive:
                raise ArgumentError("Cannot restore SensorLog, walker %s exists in restored log but did not exist before" % str(walker.selector))

    def set_rollover(self, area, enabled):
        """Configure whether rollover is enabled for streaming or storage streams.

        Normally a SensorLog is used in ring-buffer mode which means that old
        readings are automatically overwritten as needed when new data is saved.

        However, you can configure it into fill-stop mode by using:
        set_rollover("streaming"|"storage", True|False)

        By default rollover is set to True for both streaming and storage and can
        be controlled individually for each one.

        Args:
            area (str): Either streaming or storage.
            enabled (bool): Whether to enable or disable rollover.
        """

        if area == u'streaming':
            self._rollover_streaming = enabled
        elif area == u'storage':
            self._rollover_storage = enabled
        else:
            raise ArgumentError("You must pass one of 'storage' or 'streaming' to set_rollover", area=area)

    def dump_constants(self):
        """Dump (stream, value) pairs for all constant streams.

        This method walks the internal list of defined stream walkers and
        dumps the current value for all constant streams.

        Returns:
            list of (DataStream, IOTileReading): A list of all of the defined constants.
        """

        constants = []

        for walker in self._virtual_walkers:
            if not walker.selector.inexhaustible:
                continue

            constants.append((walker.selector.as_stream(), walker.reading))

        return constants

    def watch(self, selector, callback):
        """Call a function whenever a stream changes.

        Args:
            selector (DataStreamSelector): The selector to watch.
                If this is None, it is treated as a wildcard selector
                that matches every stream.
            callback (callable): The function to call when a new
                reading is pushed.  Callback is called as:
                callback(stream, value)
        """

        if selector not in self._monitors:
            self._monitors[selector] = set()

        self._monitors[selector].add(callback)

    def create_walker(self, selector, skip_all=True):
        """Create a stream walker based on the given selector.

        This function returns a StreamWalker subclass that will
        remain up to date and allow iterating over and popping readings
        from the stream(s) specified by the selector.

        When the stream walker is done, it should be passed to
        destroy_walker so that it is removed from internal lists that
        are used to always keep it in sync.

        Args:
            selector (DataStreamSelector): The selector describing the
                streams that we want to iterate over.
            skip_all (bool): Whether to start at the beginning of the data
                or to skip everything and start at the end.  Defaults
                to skipping everything.  This parameter only has any
                effect on buffered stream selectors.

        Returns:
            StreamWalker: A properly updating stream walker with the given selector.
        """

        if selector.buffered:
            walker = BufferedStreamWalker(selector, self._engine, skip_all=skip_all)
            self._queue_walkers.append(walker)
            return walker

        if selector.match_type == DataStream.CounterType:
            walker = CounterStreamWalker(selector)
        else:
            walker = VirtualStreamWalker(selector)

        self._virtual_walkers.append(walker)

        return walker

    def destroy_walker(self, walker):
        """Destroy a previously created stream walker.

        Args:
            walker (StreamWalker): The walker to remove from internal updating
                lists.
        """

        if walker.buffered:
            self._queue_walkers.remove(walker)
        else:
            self._virtual_walkers.remove(walker)

    def restore_walker(self, dumped_state):
        """Restore a stream walker that was previously serialized.

        Since stream walkers need to be tracked in an internal list for
        notification purposes, we need to be careful with how we restore
        them to make sure they remain part of the right list.

        Args:
            dumped_state (dict): The dumped state of a stream walker
                from a previous call to StreamWalker.dump()

        Returns:
            StreamWalker: The correctly restored StreamWalker subclass.
        """

        selector_string = dumped_state.get(u'selector')
        if selector_string is None:
            raise ArgumentError("Invalid stream walker state in restore_walker, missing 'selector' key", state=dumped_state)

        selector = DataStreamSelector.FromString(selector_string)

        walker = self.create_walker(selector)
        walker.restore(dumped_state)
        return walker

    def destroy_all_walkers(self):
        """Destroy any previously created stream walkers."""

        self._queue_walkers = []
        self._virtual_walkers = []

    def count(self):
        """Count many many readings are persistently stored.

        Returns:
            (int, int): The number of readings in storage and output areas.
        """

        return self._engine.count()

    def clear(self):
        """Clear all data from this sensor_log.

        All readings in all walkers are skipped and buffered data is
        destroyed.
        """

        for walker in self._virtual_walkers:
            walker.skip_all()

        self._engine.clear()

        for walker in self._queue_walkers:
            walker.skip_all()

        self._last_values = {}

    def push(self, stream, reading):
        """Push a reading into a stream, updating any associated stream walkers.

        Args:
            stream (DataStream): the stream to push the reading into
            reading (IOTileReading): the reading to push
        """

        # Make sure the stream is correct
        reading = copy.copy(reading)
        reading.stream = stream.encode()

        if stream.buffered:
            output_buffer = stream.output

            if self.id_assigner is not None:
                reading.reading_id = self.id_assigner(stream, reading)

            try:
                self._engine.push(reading)
            except StorageFullError:
                # If we are in fill-stop mode, don't auto erase old data.
                if (stream.output and not self._rollover_streaming) or (not stream.output and not self._rollover_storage):
                    raise

                self._erase_buffer(stream.output)
                self._engine.push(reading)

            for walker in self._queue_walkers:
                # Only notify the walkers that are on this queue
                if walker.selector.output == output_buffer:
                    walker.notify_added(stream)

        # Activate any monitors we have for this stream
        for selector in self._monitors:
            if selector is None or selector.matches(stream):
                for callback in self._monitors[selector]:
                    callback(stream, reading)

        # Virtual streams live only in their walkers, so update each walker
        # that contains this stream.
        for walker in self._virtual_walkers:
            if walker.matches(stream):
                walker.push(stream, reading)

        self._last_values[stream] = reading

    def _erase_buffer(self, output_buffer):
        """Erase readings in the specified buffer to make space."""

        erase_size = self._model.get(u'buffer_erase_size')

        buffer_type = u'storage'
        if output_buffer:
            buffer_type = u'streaming'

        old_readings = self._engine.popn(buffer_type, erase_size)

        # Now go through all of our walkers that could match and
        # update their availability counts and data buffer pointers
        for reading in old_readings:
            stream = DataStream.FromEncoded(reading.stream)

            for walker in self._queue_walkers:
                # Only notify the walkers that are on this queue
                if walker.selector.output == output_buffer:
                    walker.notify_rollover(stream)

    def inspect_last(self, stream, only_allocated=False):
        """Return the last value pushed into a stream.

        This function works even if the stream is virtual and no
        virtual walker has been created for it.  It is primarily
        useful to aid in debugging sensor graphs.

        Args:
            stream (DataStream): The stream to inspect.
            only_allocated (bool): Optional parameter to only allow inspection
                of allocated virtual streams.  This is useful for mimicking the
                behavior of an embedded device that does not have a _last_values
                array.

        Returns:
            IOTileReading: The data in the stream

        Raises:
            StreamEmptyError: if there has never been data written to
                the stream.
            UnresolvedIdentifierError: if only_allocated is True and there has not
                been a virtual stream walker allocated to listen to this stream.
        """

        if only_allocated:
            found = False
            for walker in self._virtual_walkers:
                if walker.matches(stream):
                    found = True
                    break

            if not found:
                raise UnresolvedIdentifierError("inspect_last could not find an allocated virtual streamer for the desired stream", stream=stream)

        if stream in self._last_values:
            return self._last_values[stream]

        raise StreamEmptyError(u"inspect_last called on stream that has never been written to", stream=stream)



"""Classes for representing and selecting data streams.

Streams are basically FIFOs with a numeric tag to represent their
identity and a series of readings.  There are two key classes:

DataStream: A wrapper around the tag that identifies the stream

DataStreamSelector: An object that selects a specific stream or a
    class of streams.
"""

# For python 2/3 compatibility using future module
from builtins import str
from future.utils import python_2_unicode_compatible, iteritems
from iotile.core.exceptions import ArgumentError, InternalError


@python_2_unicode_compatible
class DataStream(object):
    """An immutable specifier of a specific data stream

    Args:
        stream_type (int): The type of the stream
        stream_id (int): The stream identifier
        system (bool): Whether this stream is user defined or
            has global, system-wide meaning.
    """

    ImportantSystemOutputStart = 1024
    ImportantSystemStorageStart = 1024 + 512

    BufferedType = 0
    UnbufferedType = 1
    ConstantType = 2
    InputType = 3
    CounterType = 4
    OutputType = 5

    StringToType = {
        u'buffered': 0,
        u'unbuffered': 1,
        u'constant': 2,
        u'input': 3,
        u'counter': 4,
        u'output': 5
    }

    TypeToString = {
        0: u'buffered',
        1: u'unbuffered',
        2: u'constant',
        3: u'input',
        4: u'counter',
        5: u'output'
    }

    # These streams are globally important system stream codes that
    # are included by default in all wildcard stream matching.
    KnownBreakStreams = {
        1024: 'device_reboot'
    }

    def __init__(self, stream_type, stream_id, system=False):
        self.stream_type = stream_type
        self.stream_id = stream_id
        self.system = system

    @property
    def input(self):
        return self.stream_type == DataStream.InputType

    @property
    def buffered(self):
        return self.stream_type == self.BufferedType or self.stream_type == self.OutputType

    @property
    def output(self):
        return self.stream_type == DataStream.OutputType

    @property
    def important(self):
        return self.stream_type == DataStream.InputType and self.system and self.stream_id >= DataStream.ImportantSystemOutputStart

    def associated_stream(self):
        """Return the corresponding output or storage stream for an important system input.

        Certain system inputs are designed as important and automatically
        copied to output streams without requiring any manual interaction.

        This method returns the corresponding stream for an important system
        input.  It will raise an InternalError unlesss the self.important
        property is True.

        Returns:
            DataStream: The corresponding output or storage stream.

        Raises:
            InternalError: If this stream is not marked as an important system input.
        """

        if not self.important:
            raise InternalError("You may only call autocopied_stream on when DataStream.important is True", stream=self)

        if self.stream_id >= DataStream.ImportantSystemStorageStart:
            stream_type = DataStream.BufferedType
        else:
            stream_type = DataStream.OutputType

        return DataStream(stream_type, self.stream_id, True)

    @classmethod
    def FromString(cls, string_rep):
        """Create a DataStream from a string representation.

        The format for stream designators when encoded as strings is:
        [system] (buffered|unbuffered|constant|input|count|output) <integer>

        Args:
            string_rep (str): The string representation to turn into a
                DataStream
        """

        rep = str(string_rep)

        parts = rep.split()
        if len(parts) > 3:
            raise ArgumentError("Too many whitespace separated parts of stream designator", input_string=string_rep)
        elif len(parts) == 3 and parts[0] != u'system':
            raise ArgumentError("Too many whitespace separated parts of stream designator", input_string=string_rep)
        elif len(parts) < 2:
            raise ArgumentError("Too few components in stream designator", input_string=string_rep)

        # Now actually parse the string
        if len(parts) == 3:
            system = True
            stream_type = parts[1]
            stream_id = parts[2]
        else:
            system = False
            stream_type = parts[0]
            stream_id = parts[1]

        try:
            stream_id = int(stream_id, 0)
        except ValueError as exc:
            raise ArgumentError("Could not convert stream id to integer", error_string=str(exc), stream_id=stream_id)

        try:
            stream_type = cls.StringToType[stream_type]
        except KeyError:
            raise ArgumentError("Invalid stream type given", stream_type=stream_type, known_types=cls.StringToType.keys())

        return DataStream(stream_type, stream_id, system)

    @classmethod
    def FromEncoded(self, encoded):
        """Create a DataStream from an encoded 16-bit unsigned integer.

        Returns:
            DataStream: The decoded DataStream object
        """

        stream_type = (encoded >> 12) & 0b1111
        stream_system = bool(encoded & (1 << 11))
        stream_id = (encoded & ((1 << 11) - 1))

        return DataStream(stream_type, stream_id, stream_system)

    def encode(self):
        """Encode this stream as a packed 16-bit unsigned integer.

        Returns:
            int: The packed encoded stream
        """

        return (self.stream_type << 12) | (int(self.system) << 11) | self.stream_id

    def __str__(self):
        type_str = self.TypeToString[self.stream_type]

        if self.system:
            return u'system {} {}'.format(type_str, self.stream_id)

        return u'{} {}'.format(type_str, self.stream_id)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if not isinstance(other, DataStream):
            return NotImplemented

        return self.system == other.system and self.stream_type == other.stream_type and self.stream_id == other.stream_id

    def __ne__(self, other):
        if not isinstance(other, DataStream):
            return NotImplemented

        return not self == other


@python_2_unicode_compatible
class DataStreamSelector(object):
    """A specifier that matches DataStreams based on their types and ids.

    Args:
        stream_type (int): The type of the stream to match
        stream_id (int): The stream identifier to match.  If None, all
            streams of the given type are matched.
        stream_specifier (int): One of the 4 below match specifiers.
            These control whether user and/or system values are matched.
    """

    MatchSystemOnly = 1
    MatchCombined = 2
    MatchUserOnly = 3
    MatchUserAndBreaks = 4

    ValidSpecifiers = frozenset([MatchSystemOnly, MatchCombined, MatchUserOnly, MatchUserAndBreaks])

    SpecifierStrings = {
        MatchSystemOnly: u'system',
        MatchCombined: u'combined',
        MatchUserOnly: u'user',
        MatchUserAndBreaks: u''
    }

    SpecifierNames = {y: x for x, y in iteritems(SpecifierStrings)}

    SpecifierEncodings = {
        MatchSystemOnly: (1 << 11),
        MatchUserOnly: 0,
        MatchUserAndBreaks: (1 << 15),
        MatchCombined: (1 << 11) | (1 << 15)
    }

    SpecifierEncodingMap = {y: x for x, y in iteritems(SpecifierEncodings)}
    MatchAllCode = (1 << 11) - 1

    def __init__(self, stream_type, stream_id, stream_specifier):
        self.match_type = stream_type
        self.match_id = stream_id
        self.match_spec = stream_specifier

        if stream_specifier not in DataStreamSelector.ValidSpecifiers:
            raise ArgumentError("Unknown stream selector specifier", specifier=stream_specifier, known_specifiers=DataStreamSelector.ValidSpecifiers)

    @property
    def input(self):
        """Whether this is a root input stream."""

        return self.match_type == DataStream.InputType

    @property
    def inexhaustible(self):
        """Whether this is a constant stream."""

        return self.match_type == DataStream.ConstantType

    @property
    def singular(self):
        """Whether this selector matches only a single stream."""

        return self.match_id is not None

    @property
    def buffered(self):
        return self.match_type == DataStream.BufferedType or self.match_type == DataStream.OutputType

    @property
    def output(self):
        return self.match_type == DataStream.OutputType

    def as_stream(self):
        """Convert this selector to a DataStream.

        This function will only work if this is a singular selector that
        matches exactly one DataStream.
        """

        if not self.singular:
            raise ArgumentError("Attempted to convert a non-singular selector to a data stream, it matches multiple", selector=self)

        return DataStream(self.match_type, self.match_id, self.match_spec == DataStreamSelector.MatchSystemOnly)

    @classmethod
    def FromStream(cls, stream):
        """Create a DataStreamSelector from a DataStream.

        Args:
            stream (DataStream): The data stream that we want to convert.
        """

        if stream.system:
            specifier = DataStreamSelector.MatchSystemOnly
        else:
            specifier = DataStreamSelector.MatchUserOnly

        return DataStreamSelector(stream.stream_type, stream.stream_id, specifier)

    @classmethod
    def FromEncoded(cls, encoded):
        """Create a DataStreamSelector from an encoded 16-bit value.

        The binary value must be equivalent to what is produced by
        a call to self.encode() and will turn that value back into
        a a DataStreamSelector.

        Note that the following operation is a no-op:

        DataStreamSelector.FromEncode(value).encode()

        Args:
            encoded (int): The encoded binary representation of a
                DataStreamSelector.

        Returns:
            DataStreamSelector: The decoded selector.
        """

        match_spec = encoded & ((1 << 11) | (1 << 15))
        match_type = (encoded & (0b111 << 12)) >> 12
        match_id = encoded & ((1 << 11) - 1)

        if match_spec not in cls.SpecifierEncodingMap:
            raise ArgumentError("Unknown encoded match specifier", match_spec=match_spec, known_specifiers=cls.SpecifierEncodingMap.keys())

        spec_name = cls.SpecifierEncodingMap[match_spec]

        # Handle wildcard matches
        if match_id == cls.MatchAllCode:
            match_id = None

        return DataStreamSelector(match_type, match_id, spec_name)

    @classmethod
    def FromString(cls, string_rep):
        """Create a DataStreamSelector from a string.

        The format of the string should either be:

        all <type>
        OR
        <type> <id>

        Where type is [system] <stream type>, with <stream type>
        defined as in DataStream

        Args:
            rep (str): The string representation to convert to a DataStreamSelector
        """

        rep = str(string_rep)

        rep = rep.replace(u'node', '')
        rep = rep.replace(u'nodes', '')

        if rep.startswith(u'all'):
            parts = rep.split()

            spec_string = u''

            if len(parts) == 3:
                spec_string = parts[1]
                stream_type = parts[2]
            elif len(parts) == 2:
                stream_type = parts[1]
            else:
                raise ArgumentError("Invalid wildcard stream selector", string_rep=string_rep)

            try:
                # Remove pluralization that can come with e.g. 'all system outputs'
                if stream_type.endswith(u's'):
                    stream_type = stream_type[:-1]

                stream_type = DataStream.StringToType[stream_type]
            except KeyError:
                raise ArgumentError("Invalid stream type given", stream_type=stream_type, known_types=DataStream.StringToType.keys())

            stream_spec = DataStreamSelector.SpecifierNames.get(spec_string, None)
            if stream_spec is None:
                raise ArgumentError("Invalid stream specifier given (should be system, user, combined or blank)", string_rep=string_rep, spec_string=spec_string)

            return DataStreamSelector(stream_type, None, stream_spec)

        # If we're not matching a wildcard stream type, then the match is exactly
        # the same as a DataStream identifier, so use that to match it.

        stream = DataStream.FromString(rep)
        return DataStreamSelector.FromStream(stream)

    def __str__(self):
        type_str = DataStream.TypeToString[self.match_type]

        if self.match_id is not None:
            if self.match_spec == DataStreamSelector.MatchSystemOnly:
                return u'system {} {}'.format(type_str, self.match_id)

            return u'{} {}'.format(type_str, self.match_id)
        else:
            specifier = DataStreamSelector.SpecifierStrings[self.match_spec]

            if specifier != u'':
                specifier += u' '

            if type_str != u'buffered':
                type_str = type_str + 's'

            return u'all {}{}'.format(specifier, type_str)

    def matches(self, stream):
        """Check if this selector matches the given stream

        Args:
            stream (DataStream): The stream to check

        Returns:
            bool: True if this selector matches the stream
        """

        if self.match_type != stream.stream_type:
            return False

        if self.match_id is not None:
            return self.match_id == stream.stream_id

        if self.match_spec == DataStreamSelector.MatchUserOnly:
            return not stream.system
        elif self.match_spec == DataStreamSelector.MatchSystemOnly:
            return stream.system
        elif self.match_spec == DataStreamSelector.MatchUserAndBreaks:
            return (not stream.system) or (stream.system and (stream.stream_id in DataStream.KnownBreakStreams))

        # The other case is that match_spec is MatchCombined, which matches everything
        # regardless of system of user flag
        return True

    def encode(self):
        """Encode this stream as a packed 16-bit unsigned integer.

        Returns:
            int: The packed encoded stream
        """

        match_id = self.match_id
        if match_id is None:
            match_id = (1 << 11) - 1

        return (self.match_type << 12) | DataStreamSelector.SpecifierEncodings[self.match_spec] | match_id

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if not isinstance(other, DataStreamSelector):
            return NotImplemented

        return self.match_spec == other.match_spec and self.match_type == other.match_type and self.match_id == other.match_id

    def __ne__(self, other):
        if not isinstance(other, DataStreamSelector):
            return NotImplemented

        return not self == other



"""Sensor Graph main object."""

from collections import deque
import logging
from pkg_resources import iter_entry_points
from toposort import toposort_flatten
from iotile.core.exceptions import ArgumentError
from iotile.core.hw.reports import IOTileReading
from .node_descriptor import parse_node_descriptor
from .slot import SlotIdentifier
from .stream import DataStream
from .known_constants import config_fast_tick_secs, config_tick1_secs, config_tick2_secs
from .exceptions import NodeConnectionError, ProcessingFunctionError, ResourceUsageError


class SensorGraph(object):
    """A graph based data processing engine.

    Args:
        sensor_log (SensorLog): The sensor log where we should store our
            data
        model (DeviceModel): The device model to use for limiting our
            available resources
        enforce_limits (bool): Enforce the sensor graph size limits imposed
            by the chosen device model.  This can be useful for getting early
            failures on sensor graphs that cannot work on a given device model.
            Defaults to False.
    """

    def __init__(self, sensor_log, model=None, enforce_limits=False):
        self.roots = []
        self.nodes = []
        self.streamers = []

        self.constant_database = {}
        self.metadata_database = {}
        self.config_database = {}

        self.sensor_log = sensor_log
        self.model = model

        self._manually_triggered_streamers = set()
        self._logger = logging.getLogger(__name__)

        if enforce_limits:
            if model is None:
                raise ArgumentError("You must pass a device model if you set enforce_limits=True")

            self._max_nodes = model.get(u'max_nodes')
            self._max_streamers = model.get(u'max_streamers')
        else:
            self._max_nodes = None
            self._max_streamers = None

    def clear(self):
        """Clear all nodes from this sensor_graph.

        This function is equivalent to just creating a new SensorGraph() object
        from scratch.  It does not clear any data from the SensorLog, however.
        """

        self.roots = []
        self.nodes = []
        self.streamers = []

        self.constant_database = {}
        self.metadata_database = {}
        self.config_database = {}

    def add_node(self, node_descriptor):
        """Add a node to the sensor graph based on the description given.

        The node_descriptor must follow the sensor graph DSL and describe
        a node whose input nodes already exist.

        Args:
            node_descriptor (str): A description of the node to be added
                including its inputs, triggering conditions, processing function
                and output stream.
        """

        if self._max_nodes is not None and len(self.nodes) >= self._max_nodes:
            raise ResourceUsageError("Maximum number of nodes exceeded", max_nodes=self._max_nodes)

        node, inputs, processor = parse_node_descriptor(node_descriptor, self.model)

        in_root = False

        for i, input_data in enumerate(inputs):
            selector, trigger = input_data

            walker = self.sensor_log.create_walker(selector)

            # Constant walkers begin life initialized to 0 so they always read correctly
            if walker.selector.inexhaustible:
                walker.reading = IOTileReading(walker.selector.as_stream(), 0xFFFFFFFF, 0)

            node.connect_input(i, walker, trigger)

            if selector.input and not in_root:
                self.roots.append(node)
                in_root = True  # Make sure we only add to root list once
            else:
                found = False
                for other in self.nodes:
                    if selector.matches(other.stream):
                        other.connect_output(node)
                        found = True

                if not found and selector.buffered:
                    raise NodeConnectionError("Node has input that refers to another node that has not been created yet", node_descriptor=node_descriptor, input_selector=str(selector), input_index=i)

        # Also make sure we add this node's output to any other existing node's inputs
        # this is important for constant nodes that may be written from multiple places
        # FIXME: Make sure when we emit nodes, they are topologically sorted
        for other_node in self.nodes:
            for selector, trigger in other_node.inputs:
                if selector.matches(node.stream):
                    node.connect_output(other_node)

        # Find and load the processing function for this node
        func = self.find_processing_function(processor)
        if func is None:
            raise ProcessingFunctionError("Could not find processing function in installed packages", func_name=processor)

        node.set_func(processor, func)
        self.nodes.append(node)

    def add_config(self, slot, config_id, config_type, value):
        """Add a config variable assignment to this sensor graph.

        Args:
            slot (SlotIdentifier): The slot identifier that this config
                variable is assigned to.
            config_id (int): The 16-bit id of this config_id
            config_type (str): The type of the config variable, currently
                supported are fixed width integer types, strings and binary
                blobs.
            value (str|int|bytes): The value to assign to the config variable.
        """

        if slot not in self.config_database:
            self.config_database[slot] = {}

        self.config_database[slot][config_id] = (config_type, value)

    def add_streamer(self, streamer):
        """Add a streamer to this sensor graph.

        Args:
            streamer (DataStreamer): The streamer we want to add
        """

        if self._max_streamers is not None and len(self.streamers) >= self._max_streamers:
            raise ResourceUsageError("Maximum number of streamers exceeded", max_streamers=self._max_streamers)

        streamer.link_to_storage(self.sensor_log)
        streamer.index = len(self.streamers)

        self.streamers.append(streamer)

    def add_constant(self, stream, value):
        """Store a constant value for use in this sensor graph.

        Constant assignments occur after all sensor graph nodes have been
        allocated since they must be propogated to all appropriate virtual
        stream walkers.

        Args:
            stream (DataStream): The constant stream to assign the value to
            value (int): The value to assign.
        """

        if stream in self.constant_database:
            raise ArgumentError("Attempted to set the same constant twice", stream=stream, old_value=self.constant_database[stream], new_value=value)

        self.constant_database[stream] = value

    def add_metadata(self, name, value):
        """Attach a piece of metadata to this sensorgraph.

        Metadata is not used during the simulation of a sensorgraph but allows
        it to convey additional context that may be used during code
        generation.  For example, associating an `app_tag` with a sensorgraph
        allows the snippet code generator to set that app_tag on a device when
        programming the sensorgraph.

        Arg:
            name (str): The name of the metadata that we wish to associate with this
                sensorgraph.
            value (object): The value we wish to store.
        """

        if name in self.metadata_database:
            raise ArgumentError("Attempted to set the same metadata value twice", name=name, old_value=self.metadata_database[name], new_value=value)

        self.metadata_database[name] = value

    def initialize_remaining_constants(self, value=0):
        """Ensure that all constant streams referenced in the sensor graph have a value.

        Constant streams that are automatically created by the compiler are initialized
        as part of the compilation process but it's possible that the user references
        other constant streams but never assigns them an explicit initial value.  This
        function will initialize them all to a default value (0 if not passed) and
        return the streams that were so initialized.

        Args:
            value (int): Optional value to use to initialize all uninitialized constants.
                Defaults to 0 if not passed.

        Returns:
            list(DataStream): A list of all of the constant streams that were not previously
                initialized and were initialized to the given value in this function.
        """

        remaining = []

        for node, _inputs, _outputs in self.iterate_bfs():
            streams = node.input_streams() + [node.stream]

            for stream in streams:
                if stream.stream_type is not DataStream.ConstantType:
                    continue

                if stream not in self.constant_database:
                    self.add_constant(stream, value)
                    remaining.append(stream)

        return remaining

    def load_constants(self):
        """Load all constants into their respective streams.

        All previous calls to add_constant stored a constant value that
        should be associated with virtual stream walkers.  This function
        actually calls push_stream in order to push all of the constant
        values to their walkers.
        """

        for stream, value in self.constant_database.items():
            self.sensor_log.push(stream, IOTileReading(stream.encode(), 0, value))

    def get_config(self, slot, config_id):
        """Get a config variable assignment previously set on this sensor graph.

        Args:
            slot (SlotIdentifier): The slot that we are setting this config variable
                on.
            config_id (int): The 16-bit config variable identifier.

        Returns:
            (str, str|int): Returns a tuple with the type of the config variable and
                the value that is being set.

        Raises:
            ArgumentError: If the config variable is not currently set on the specified
                slot.
        """

        if slot not in self.config_database:
            raise ArgumentError("No config variables have been set on specified slot", slot=slot)

        if config_id not in self.config_database[slot]:
            raise ArgumentError("Config variable has not been set on specified slot", slot=slot, config_id=config_id)

        return self.config_database[slot][config_id]

    def is_output(self, stream):
        """Check if a stream is a sensor graph output.

        Return:
            bool
        """

        for streamer in self.streamers:
            if streamer.selector.matches(stream):
                return True

        return False

    def get_tick(self, name):
        """Check the config variables to see if there is a configurable tick.

        Sensor Graph has a built-in 10 second tick that is sent every 10
        seconds to allow for triggering timed events.  There is a second
        'user' tick that is generated internally by the sensorgraph compiler
        and used for fast operations and finally there are several field
        configurable ticks that can be used for setting up configurable
        timers.

        This is done by setting a config variable on the controller with the
        desired tick interval, which is then interpreted by this function.

        The appropriate config_id to use is listed in `known_constants.py`

        Returns:
            int: 0 if the tick is disabled, otherwise the number of seconds
                between each tick
        """

        name_map = {
            'fast': config_fast_tick_secs,
            'user1': config_tick1_secs,
            'user2': config_tick2_secs
        }

        config = name_map.get(name)
        if config is None:
            raise ArgumentError("Unknown tick requested", name=name)

        slot = SlotIdentifier.FromString('controller')

        try:
            var = self.get_config(slot, config)
            return var[1]
        except ArgumentError:
            return 0

    def process_input(self, stream, value, rpc_executor):
        """Process an input through this sensor graph.

        The tick information in value should be correct and is transfered
        to all results produced by nodes acting on this tick.

        Args:
            stream (DataStream): The stream the input is part of
            value (IOTileReading): The value to process
            rpc_executor (RPCExecutor): An object capable of executing RPCs
                in case we need to do that.
        """

        self.sensor_log.push(stream, value)

        # FIXME: This should be specified in our device model
        if stream.important:
            associated_output = stream.associated_stream()
            self.sensor_log.push(associated_output, value)

        to_check = deque([x for x in self.roots])

        while len(to_check) > 0:
            node = to_check.popleft()
            if node.triggered():
                try:
                    results = node.process(rpc_executor, self.mark_streamer)
                    for result in results:
                        result.raw_time = value.raw_time
                        self.sensor_log.push(node.stream, result)
                except:
                    self._logger.exception("Unhandled exception in graph node processing function for node %s", str(node))

                # If we generated any outputs, notify our downstream nodes
                # so that they are also checked to see if they should run.
                if len(results) > 0:
                    to_check.extend(node.outputs)

    def mark_streamer(self, index):
        """Manually mark a streamer that should trigger.

        The next time check_streamers is called, the given streamer will be
        manually marked that it should trigger, which will cause it to trigger
        unless it has no data.

        Args:
            index (int): The index of the streamer that we should mark as
                manually triggered.

        Raises:
            ArgumentError: If the streamer index is invalid.
        """

        self._logger.debug("Marking streamer %d manually", index)
        if index >= len(self.streamers):
            raise ArgumentError("Invalid streamer index", index=index, num_streamers=len(self.streamers))

        self._manually_triggered_streamers.add(index)

    def check_streamers(self, blacklist=None):
        """Check if any streamers are ready to produce a report.

        You can limit what streamers are checked by passing a set-like
        object into blacklist.

        This method is the primary way to see when you should poll a given
        streamer for its next report.

        Note, this function is not idempotent.  If a streamer is marked as
        manual and it is triggered from a node rule inside the sensor_graph,
        that trigger will only last as long as the next call to
        check_streamers() so you need to explicitly build a report on all
        ready streamers before calling check_streamers again.

        Args:
            blacklist (set): Optional set of streamer indices that should
                not be checked right now.

        Returns:
            list of DataStreamer: A list of the ready streamers.
        """

        ready = []
        selected = set()

        for i, streamer in enumerate(self.streamers):
            if blacklist is not None and i in blacklist:
                continue

            if i in selected:
                continue

            marked = False
            if i in self._manually_triggered_streamers:
                marked = True
                self._manually_triggered_streamers.remove(i)

            if streamer.triggered(marked):
                self._logger.debug("Streamer %d triggered, manual=%s", i, marked)
                ready.append(streamer)
                selected.add(i)

                # Handle streamers triggered with another
                for j, streamer2 in enumerate(self.streamers[i:]):
                    if streamer2.with_other == i and j not in selected and streamer2.triggered(True):
                        self._logger.debug("Streamer %d triggered due to with-other on %d", j, i)
                        ready.append(streamer2)
                        selected.add(j)

        return ready

    def iterate_bfs(self):
        """Generator that yields node, [inputs], [outputs] in breadth first order.

        This generator will iterate over all nodes in the sensor graph, yielding
        a 3 tuple for each node with a list of all of the nodes connected to its
        inputs and all of the nodes connected to its output.

        Returns:
            (SGNode, list(SGNode), list(SGNode)): A tuple for each node in the graph
        """

        working_set = deque(self.roots)
        seen = []

        while len(working_set) > 0:
            curr = working_set.popleft()

            # Now build input and output node lists for this node
            inputs = []
            for walker, _ in curr.inputs:
                for other in seen:
                    if walker.matches(other.stream) and other not in inputs:
                        inputs.append(other)

            outputs = [x for x in curr.outputs]
            yield curr, inputs, outputs

            working_set.extend(curr.outputs)
            seen.append(curr)

    def sort_nodes(self):
        """Topologically sort all of our nodes.

        Topologically sorting our nodes makes nodes that are inputs to other
        nodes come first in the list of nodes.  This is important to do before
        programming a sensorgraph into an embedded device whose engine assumes
        a topologically sorted graph.

        The sorting is done in place on self.nodes
        """

        node_map = {id(node): i for i, node in enumerate(self.nodes)}
        node_deps = {}

        for node, inputs, _outputs in self.iterate_bfs():
            node_index = node_map[id(node)]

            deps = {node_map[id(x)] for x in inputs}
            node_deps[node_index] = deps

        # Now that we have our dependency tree properly built, topologically
        # sort the nodes and reorder them.
        node_order = toposort_flatten(node_deps)
        self.nodes = [self.nodes[x] for x in node_order]

        #Check root nodes all topographically sorted to the beginning
        for root in self.roots:
            if root not in self.nodes[0:len(self.roots)]:
                raise NodeConnectionError("Inputs not sorted in the beginning", node=str(root), node_position=self.nodes.index(root))

    @classmethod
    def find_processing_function(cls, name):
        """Find a processing function by name.

        This function searches through installed processing functions
        using pkg_resources.

        Args:
            name (str): The name of the function we're looking for

        Returns:
            callable: The processing function
        """

        for entry in iter_entry_points(u'iotile.sg_processor', name):
            return entry.load()

    def dump_nodes(self):
        """Dump all of the nodes in this sensor graph as a list of strings."""

        return [str(x) for x in self.nodes]

    def dump_streamers(self):
        """Dump all of the streamers in this sensor graph as a list of strings."""

        return [str(streamer) for streamer in self.streamers]
